"""Test import existing project samples command."""

# pylint: disable=wrong-import-order, import-error
import operator
from unittest import mock
from uuid import uuid4

from click.testing import CliRunner

from gencove.client import APIClient, APIClientError
from gencove.command.projects.cli import import_existing_project_samples
from gencove.constants import IMPORT_BATCH_SIZE
from gencove.models import (
    ImportExistingSamplesModel,
    ProjectSamples,
    SampleImport,
)
from gencove.tests.decorators import assert_authorization, assert_no_requests
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.projects.vcr.filters import (
    filter_get_project_samples_request,
    filter_get_project_samples_response,
    filter_import_existing_samples_request,
    filter_import_existing_samples_response,
)
from gencove.tests.utils import MOCK_UUID, get_vcr_response
from gencove.utils import batchify

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_get_project_samples_request,
            filter_import_existing_samples_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_get_project_samples_response,
            filter_import_existing_samples_response,
        ],
    }


@assert_no_requests
def test_import_existing_project_samples__bad_project_id(mocker, credentials):
    """Test import existing project samples failure when non-uuid string is used as
    project id.
    """
    runner = CliRunner()

    mocked_import_existing_samples = mocker.patch.object(
        APIClient,
        "import_existing_samples",
    )

    res = runner.invoke(
        import_existing_project_samples,
        [
            "1111111",
            "--sample-ids",
            f"{str(uuid4())},{str(uuid4())}",
            *credentials,
        ],
    )

    assert res.exit_code == 1
    mocked_import_existing_samples.assert_not_called()
    assert "Project ID is not valid" in res.output


@assert_no_requests
def test_import_existing_project_samples__bad_sample_ids(mocker, credentials):
    """Test import existing project samples failure when sample_ids are bad."""
    runner = CliRunner()

    mocked_import_existing_samples = mocker.patch.object(
        APIClient,
        "import_existing_samples",
    )

    res = runner.invoke(
        import_existing_project_samples,
        [
            str(uuid4()),
            "--sample-ids",
            "1111111,2222222",
            *credentials,
        ],
    )

    assert res.exit_code == 1
    mocked_import_existing_samples.assert_not_called()
    assert "Not all source sample IDs are valid" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_import_existing_project_samples__bad_source_project_id(mocker, credentials):
    """Test import existing project samples failure when non-uuid string is used as
    source project id.
    """
    runner = CliRunner()

    mocked_import_existing_samples = mocker.patch.object(
        APIClient,
        "import_existing_samples",
    )

    res = runner.invoke(
        import_existing_project_samples,
        [
            str(uuid4()),
            "--source-project-id",
            "1111111",
            *credentials,
        ],
    )

    assert res.exit_code == 1
    mocked_import_existing_samples.assert_not_called()
    print(res.output)
    assert "Source project ID is not valid" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_import_existing_project_samples__sample_project_id(mocker, credentials):
    """Test import existing project samples failure when sampe project id is used
    as source and destination.
    """
    runner = CliRunner()

    mocked_import_existing_samples = mocker.patch.object(
        APIClient,
        "import_existing_samples",
    )

    project_id = str(uuid4())
    res = runner.invoke(
        import_existing_project_samples,
        [
            project_id,
            "--source-project-id",
            project_id,
            *credentials,
        ],
    )

    assert res.exit_code == 1
    mocked_import_existing_samples.assert_not_called()
    print(res.output)
    assert "Source and destination project must be different" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_import_existing_project_samples__both_sample_ids_and_source_project_id(
    mocker, credentials
):
    """Test import existing project samples failure when both sample ids and
    source project id are given.
    """
    runner = CliRunner()

    mocked_import_existing_samples = mocker.patch.object(
        APIClient,
        "import_existing_samples",
    )

    res = runner.invoke(
        import_existing_project_samples,
        [
            str(uuid4()),
            "--source-project-id",
            str(uuid4()),
            "--sample-ids",
            str(uuid4()),
            *credentials,
        ],
    )

    assert res.exit_code == 1
    mocked_import_existing_samples.assert_not_called()
    print(res.output)
    assert "Either --source-project-id or --sample-ids" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_import_existing_project_samples__bad_metadata(mocker, credentials):
    """Test import existing project samples failure when optional
    metadata-json is passed, but it has a bad value.
    """
    runner = CliRunner()

    mocked_import_existing_samples = mocker.patch.object(
        APIClient,
        "import_existing_samples",
    )

    res = runner.invoke(
        import_existing_project_samples,
        [
            str(uuid4()),
            "--sample-ids",
            f"{str(uuid4())},{str(uuid4())}",
            "--metadata-json",
            '[{"foo": "bar"}',
            *credentials,
        ],
    )

    assert res.exit_code == 1
    mocked_import_existing_samples.assert_not_called()
    assert "Metadata JSON is not valid" in res.output


@pytest.mark.vcr
@assert_authorization
def test_import_existing_project_samples__server_rejects(
    mocker, credentials, project_id, recording, vcr
):
    """Test import existing project samples failure when the validation
    fails on server.
    """
    runner = CliRunner()

    if not recording:
        # Mock import_existing_samples only if using the cassettes, since we
        # mock the return value.
        project_samples_import_response = get_vcr_response(
            "/api/v2/project-samples-import/",
            vcr,
            operator.contains,
            just_body=False,
        )
        mocked_import_existing_samples = mocker.patch.object(
            APIClient,
            "import_existing_samples",
            side_effect=APIClientError(
                message=project_samples_import_response["body"]["string"],
                status_code=project_samples_import_response["status"]["code"],
            ),
        )

    res = runner.invoke(
        import_existing_project_samples,
        [
            project_id,
            "--sample-ids",
            MOCK_UUID,
            *credentials,
        ],
    )
    assert res.exit_code == 1
    if not recording:
        mocked_import_existing_samples.assert_called_once()
    assert "There was an error importing" in res.output


@pytest.mark.vcr
@assert_authorization
def test_import_existing_project_samples__success(
    mocker, credentials, project_id, sample_id_import_existing, recording, vcr
):  # pylint: disable=too-many-arguments
    """Test import existing project samples success."""
    runner = CliRunner()

    if not recording:
        # Mock import_existing_samples only if using the cassettes, since we
        # mock the return value.
        project_samples_import_response = get_vcr_response(
            "/api/v2/project-samples-import/", vcr, operator.contains
        )
        mocked_import_existing_samples = mocker.patch.object(
            APIClient,
            "import_existing_samples",
            return_value=ImportExistingSamplesModel(**project_samples_import_response),
        )

    res = runner.invoke(
        import_existing_project_samples,
        [
            project_id,
            "--sample-ids",
            sample_id_import_existing,
            *credentials,
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_import_existing_samples.assert_called_once()
    assert "Number of samples imported into the project" in res.output


@pytest.mark.vcr
@assert_authorization
def test_import_existing_project__from_source_project(
    mocker, credentials, project_id_import, project_id_batches, recording, vcr
):  # pylint: disable=too-many-arguments
    """Test import existing project samples success
    Using project_id_import as destination since configuration is Import FASTQs
    Using project_id_batches as source since it only contains 2 samples
    """
    runner = CliRunner()

    if not recording:
        # Mock import_existing_samples only if using the cassettes, since we
        # mock the return value.
        mocked_samples = get_vcr_response(
            "/api/v2/project-samples/", vcr, operator.contains
        )
        # Changing "mock status" to valid values
        mocked_samples["results"][0]["last_status"]["status"] = "succeeded"
        mocked_samples["results"][1]["last_status"]["status"] = "failed qc"
        mocked_get_project_samples = mocker.patch.object(
            APIClient,
            "get_project_samples",
            return_value=ProjectSamples(**mocked_samples),
        )
        project_samples_import_response = get_vcr_response(
            "/api/v2/project-samples-import/", vcr, operator.contains
        )
        mocked_import_existing_samples = mocker.patch.object(
            APIClient,
            "import_existing_samples",
            return_value=ImportExistingSamplesModel(**project_samples_import_response),
        )

    res = runner.invoke(
        import_existing_project_samples,
        [
            project_id_import,
            "--source-project-id",
            project_id_batches,
            *credentials,
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_get_project_samples.assert_called_once()
        mocked_import_existing_samples.assert_called_once_with(
            project_id_import, [MOCK_UUID, MOCK_UUID], None
        )
    assert "Number of samples imported into the project" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_import_existing_project_samples__batch_size(
    mocker, credentials, project_id, recording, vcr  # pylint: disable=unused-argument
):  # pylint: disable=too-many-arguments
    """Test import existing project samples confirm batch size."""
    runner = CliRunner()

    sample_ids = [str(uuid4()) for _ in range(IMPORT_BATCH_SIZE * 2 + 1)]

    mocked_import_existing_samples = mocker.patch.object(
        APIClient,
        "import_existing_samples",
        return_value=ImportExistingSamplesModel(
            project_id=project_id,
            samples=[SampleImport(sample_id=sample_id) for sample_id in sample_ids],
            metadata=None,
        ),
    )

    res = runner.invoke(
        import_existing_project_samples,
        [
            project_id,
            "--sample-ids",
            ",".join(sample_ids),
            *credentials,
        ],
    )
    assert res.exit_code == 0
    assert mocked_import_existing_samples.call_count == 3
    mocked_import_existing_samples.assert_has_calls(
        [
            mock.call(project_id, samples_batch, None)
            for samples_batch in batchify(sample_ids, IMPORT_BATCH_SIZE)
        ]
    )
    assert "Number of samples imported into the project" in res.output



"""Project QC report executor."""
from pathlib import Path

from gencove import client  # noqa: I100
from gencove.command.base import Command
from gencove.command.utils import extract_filename_from_headers


class ProjectQCReport(Command):
    """Get project QC report executor."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        project_id,
        columns,
        output_filename,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.columns = columns
        self.output_filename = output_filename

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """

    def execute(self):
        """Request to get project QC report and download it."""
        self.echo_info(
            f"Retrieving project QC report for project with ID {self.project_id}"
        )

        try:
            project_qc_report = self.api_client.get_project_qc_report(
                project_id=self.project_id, columns=self.columns
            )

            if not self.output_filename:
                filename = extract_filename_from_headers(
                    headers=project_qc_report.headers
                )
                if not filename:
                    filename = "project-qc-report.csv"
                self.output_filename = Path.cwd() / filename
            else:
                Path(self.output_filename).parent.mkdir(exist_ok=True, parents=True)
            with open(self.output_filename, "wb") as filename:
                filename.write(project_qc_report.content)
            self.echo_info(
                f"Saved project QC report CSV for project {self.project_id} "
                f"to {Path(self.output_filename).resolve()}"
            )
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning(
                    "There was an error retrieving the project QC report."
                )
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            elif err.status_code == 404:
                self.echo_warning(
                    f"Project with ID {self.project_id} does not exist or "
                    "you do not have access."
                )
            else:
                raise



"""Download command exceptions."""


class DownloadTemplateError(Exception):
    """Error that indicates an issue with the template.

    Caused by download overwriting previously downloaded files.
    """



"""Configure inactivity stop for explorer instances subcommand."""

from .utils import get_line
from ....base import Command


class ListInstances(Command):
    """List explorer instances command executor."""

    def validate(self):
        """Validate list instances"""

    def initialize(self):
        """Initialize inactivity-stop subcommand."""
        self.login()

    def execute(self):
        """Make a request to list instances."""
        self.echo_debug("List instances.")

        for instance in self.api_client.get_explorer_instances().results:
            self.echo_data(get_line(instance))



"""Base command layout.

All commands must implement this interface.
"""
import contextlib
import os
import tempfile
import traceback
from functools import wraps

import click

from gencove.client import APIClient, APIClientError
from gencove.exceptions import MaintenanceError, ValidationError
from gencove.logger import (
    DEBUG,
    LOG_LEVEL,
    dump_debug_log,
    echo_data,
    echo_debug,
    echo_error,
    echo_info,
    echo_warning,
)
from gencove.utils import login, validate_credentials

AWS_PROFILE = "AWS_PROFILE"
AWS_CONFIG_FILE = "AWS_CONFIG_FILE"
AWS_SHARED_CREDENTIALS_FILE = "AWS_SHARED_CREDENTIALS_FILE"
env_var_save_restore = [
    "AWS_ACCESS_KEY_ID",
    "AWS_SECRET_ACCESS_KEY",
    "AWS_SESSION_TOKEN",
    "AWS_SECURITY_TOKEN",
    "AWS_DEFAULT_REGION",
    "AWS_DEFAULT_PROFILE",
    "AWS_CA_BUNDLE",
    AWS_SHARED_CREDENTIALS_FILE,
    AWS_CONFIG_FILE,
    AWS_PROFILE,
]

PROFILE_NAME = "gencove-cli-profile"


@contextlib.contextmanager
def aws_cli_credentials():
    """Hides all AWS environment variables, creates new AWS_PROFILE and
    associated config files. During context manager cleanup, restores all
    the environment variables and deletes the temporary files.
    """
    # Saving initial state of env variables
    env_var_saved = {}
    for env_var in env_var_save_restore:
        env_var_saved[env_var] = os.getenv(env_var, default=None)
        if env_var_saved[env_var] is not None:
            del os.environ[env_var]

    # Set AWS_PROFILE
    os.environ[AWS_PROFILE] = PROFILE_NAME

    # Set AWS_CONFIG_FILE
    with tempfile.NamedTemporaryFile(delete=False) as config_file:
        config_file.write(b"[profile " + PROFILE_NAME.encode("utf") + b"]\n")
        aws_config_file = config_file.name
        os.environ[AWS_CONFIG_FILE] = aws_config_file

    # Set AWS_SHARED_CREDENTIALS_FILE
    with tempfile.NamedTemporaryFile(delete=False) as credentials_file:
        credentials_file.write(b"[profile " + PROFILE_NAME.encode("utf") + b"]\n")
        aws_shared_credentials_file = credentials_file.name
        os.environ[AWS_SHARED_CREDENTIALS_FILE] = aws_shared_credentials_file

    try:
        yield
    finally:
        # Remove all modified env variables
        for env_var in env_var_save_restore:
            if env_var in os.environ:
                del os.environ[env_var]

        # Restore original env variables
        for env_var, env_var_value in env_var_saved.items():
            if env_var_value is not None:
                os.environ[env_var] = env_var_value

        # Remove tmp files
        os.remove(aws_config_file)
        os.remove(aws_shared_credentials_file)


def aws_cli_decorator(wrapped_func):
    """Executes a function inside the aws_cli_credentials context manager.

    Args:
        wrapped_func (callable): Function to call.
    """

    @wraps(wrapped_func)
    def wrapper(*args, **kwargs):
        with aws_cli_credentials():
            return wrapped_func(*args, **kwargs)

    return wrapper


class Command(object):  # pylint: disable=R0205
    """Base command structure."""

    def __init__(self, credentials, options):
        self.api_client = APIClient(options.host)
        self.is_logged_in = False
        self.credentials = credentials
        self.options = options
        self.fetch_credentials_from_env()
        self.is_credentials_valid = validate_credentials(credentials)

    def fetch_credentials_from_env(self):
        """Set default values for credentials from environment variables if they are
        not directly set.

        Prioritizes direct credential inputs over environment variables.

        If API key is provided, does not read email from ENV.
        Likewise, if email is provided, does not read API key from ENV.
        """
        if not any([self.credentials.api_key, self.credentials.email]):
            self.credentials.api_key = os.environ.get("GENCOVE_API_KEY", "")

        if not self.credentials.api_key:
            if not self.credentials.email:
                self.credentials.email = os.environ.get("GENCOVE_EMAIL", "")

        # Always try to set password - ignored in case of API key usage.
        if not self.credentials.password:
            self.credentials.password = os.environ.get("GENCOVE_PASSWORD", "")

    def initialize(self):
        """Put any initializing logic here, such as login."""
        raise NotImplementedError

    def validate(self):
        """Validate that the arguments and initialize were successful."""
        raise NotImplementedError

    def execute(self):
        """Execute command logic."""
        raise NotImplementedError

    def validate_login_success(self):
        """Check if login succeeded."""
        if not self.is_logged_in:
            raise ValidationError("Please check your credentials and try again")

    @aws_cli_decorator
    def run(self):
        """Run the command.

        No need to override this, unless more customized behaviour is needed.
        """
        try:
            self.initialize()
            self.validate()
            self.validate_login_success()
            self.execute()
        except ValidationError as err:
            self.echo_error(err.message)
            dump_debug_log()
            raise click.Abort()
        except MaintenanceError as err:
            self.echo_error(err.message)
            self.echo_error(f"ETA is {err.eta}")
            dump_debug_log()
            raise click.Abort()
        except APIClientError as err:
            if err.status_code == 403:
                self.echo_error(
                    "You do not have the sufficient permission "
                    "level required to perform this operation."
                )
            self.echo_error(err.message)
            dump_debug_log()
            if LOG_LEVEL == DEBUG:
                raise err
            raise click.Abort()
        except Exception as err:
            self.echo_error(traceback.format_exc())
            dump_debug_log()
            if LOG_LEVEL == DEBUG:
                raise err
            raise click.Abort()

    def login(self):
        """Login current user."""
        if not self.is_credentials_valid:
            return False
        self.is_logged_in = login(self.api_client, self.credentials)
        return self.is_logged_in

    @staticmethod
    def echo_data(msg, **kwargs):
        """Output data message."""
        echo_data(msg, **kwargs)

    @staticmethod
    def echo_info(msg, **kwargs):
        """Output info message."""
        echo_info(msg, **kwargs)

    @staticmethod
    def echo_warning(msg, **kwargs):
        """Output warning message."""
        echo_warning(msg, **kwargs)

    @staticmethod
    def echo_error(msg, **kwargs):
        """Output error message."""
        echo_error(msg, **kwargs)

    @staticmethod
    def echo_debug(msg, **kwargs):
        """Output debug message."""
        echo_debug(msg, **kwargs)
